@@ -101,6 +101,8 @@ struct _pkcs11_mech_list {
     { "CKM_DES3_MAC", CKM_DES3_MAC },
     { "CKM_DES3_MAC_GENERAL", CKM_DES3_MAC_GENERAL },
     { "CKM_DES3_CBC_PAD", CKM_DES3_CBC_PAD },
+    { "CKM_DES3_CMAC_GENERAL", CKM_DES3_CMAC_GENERAL },
+    { "CKM_DES3_CMAC", CKM_DES3_CMAC },
     { "CKM_CDMF_KEY_GEN", CKM_CDMF_KEY_GEN },
     { "CKM_CDMF_ECB", CKM_CDMF_ECB },
     { "CKM_CDMF_CBC", CKM_CDMF_CBC },
@@ -264,6 +266,8 @@ struct _pkcs11_mech_list {
     { "CKM_AES_CBC_PAD", CKM_AES_CBC_PAD },
     { "CKM_AES_CTR", CKM_AES_CTR },
     { "CKM_AES_GCM", CKM_AES_GCM },
+    { "CKM_AES_CMAC_GENERAL", CKM_AES_CMAC_GENERAL },
+    { "CKM_AES_CMAC", CKM_AES_CMAC },
     { "CKM_AES_OFB", CKM_AES_OFB },
     { "CKM_AES_CFB8", CKM_AES_CFB8 },
     { "CKM_AES_CFB64", CKM_AES_CFB64 },
@@ -272,6 +276,20 @@ struct _pkcs11_mech_list {
     { "CKM_DH_PKCS_PARAMETER_GEN", CKM_DH_PKCS_PARAMETER_GEN },
     { "CKM_X9_42_DH_PARAMETER_GEN", CKM_X9_42_DH_PARAMETER_GEN },
     { "CKM_VENDOR_DEFINED", CKM_VENDOR_DEFINED },
+    { "CKM_IBM_SHA3_224", CKM_IBM_SHA3_224 },
+    { "CKM_IBM_SHA3_256", CKM_IBM_SHA3_256 },
+    { "CKM_IBM_SHA3_384", CKM_IBM_SHA3_384 },
+    { "CKM_IBM_SHA3_512", CKM_IBM_SHA3_512 },
+    { "CKM_IBM_SHA3_224_HMAC", CKM_IBM_SHA3_224_HMAC },
+    { "CKM_IBM_SHA3_256_HMAC", CKM_IBM_SHA3_256_HMAC },
+    { "CKM_IBM_SHA3_384_HMAC", CKM_IBM_SHA3_384_HMAC },
+    { "CKM_IBM_SHA3_512_HMAC", CKM_IBM_SHA3_512_HMAC },
+    { "CKM_IBM_CMAC", CKM_IBM_CMAC },
+    { "CKM_IBM_EC_C25519", CKM_IBM_EC_C25519 },
+    { "CKM_IBM_EDDSA_SHA512", CKM_IBM_EDDSA_SHA512 },
+    { "CKM_IBM_EC_C448", CKM_IBM_EC_C448 },
+    { "CKM_IBM_ED448_SHA3", CKM_IBM_ED448_SHA3 },
+    { "CKM_IBM_DILITHIUM", CKM_IBM_DILITHIUM },
     { NULL_PTR, 0xFFFFFFFF }
 };
 
