@@ -28,6 +28,10 @@
 #include <termios.h>
 #include <errno.h>
 
+#define EP11SHAREDLIB_NAME "OCK_EP11_LIBRARY"
+#define EP11SHAREDLIB_V3 "libep11.so.3"
+#define EP11SHAREDLIB_V2 "libep11.so.2"
+#define EP11SHAREDLIB_V1 "libep11.so.1"
 #define EP11SHAREDLIB "libep11.so"
 #define PKCS11_MAX_PIN_LEN 128
 
@@ -37,25 +41,33 @@ CK_LONG adapter = -1;
 CK_LONG domain = -1;
 CK_OBJECT_HANDLE key_store[4096];
 
-typedef int (*m_get_ep11_info_t) (CK_VOID_PTR, CK_ULONG_PTR,
-                                  unsigned int, unsigned int, uint64_t);
 typedef unsigned long int (*m_admin_t) (unsigned char *, size_t *,
                                         unsigned char *,
                                         size_t *, const unsigned char *,
                                         size_t, const unsigned char *,
-                                        size_t, uint64_t);
-typedef long (*ep11a_cmdblock_t) (unsigned char *, size_t, unsigned int,
-                                  const struct ep11_admresp *,
-                                  const unsigned char *,
-                                  const unsigned char *, size_t);
-typedef long (*ep11a_internal_rv_t) (const unsigned char *, size_t,
-                                     struct ep11_admresp *, CK_RV *);
+                                        size_t, target_t);
+typedef long (*xcpa_cmdblock_t) (unsigned char *, size_t, unsigned int,
+                                 const struct XCPadmresp *,
+                                 const unsigned char *,
+                                 const unsigned char *, size_t);
+typedef long (*xcpa_internal_rv_t) (const unsigned char *, size_t,
+                                    struct XCPadmresp *, CK_RV *);
+typedef int (*m_add_module_t) (XCP_Module_t, target_t *);
+typedef int (*m_rm_module_t) (XCP_Module_t, target_t);
+typedef CK_RV (*m_get_xcp_info_t)(CK_VOID_PTR pinfo, CK_ULONG_PTR infbytes,
+                                unsigned int query, unsigned int subquery,
+                                target_t target);
 
-m_get_ep11_info_t _m_get_ep11_info;
+m_get_xcp_info_t _m_get_xcp_info;
 m_admin_t _m_admin;
-ep11a_cmdblock_t _ep11a_cmdblock;
-ep11a_internal_rv_t _ep11a_internal_rv;
+xcpa_cmdblock_t _xcpa_cmdblock;
+xcpa_internal_rv_t _xcpa_internal_rv;
+m_add_module_t _m_add_module;
+m_rm_module_t _m_rm_module;
 
+CK_VERSION lib_version;
+
+#define CK_IBM_XCPHQ_VERSION    0xff000001
 
 typedef struct {
     short format;
@@ -64,32 +76,29 @@ typedef struct {
 } __attribute__ ((packed)) ep11_target_t;
 
 
-#define blobsize 2048*4
+#define BLOBSIZE         2048*4
 
-typedef struct {
-    size_t blob_size;
-    size_t blob_id;
-    unsigned char blob[blobsize];
-} ep11_opaque;
 
-static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj, CK_BYTE *old)
+
+static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj, CK_BYTE *old,
+                     CK_ULONG old_len)
 {
-    CK_BYTE req[blobsize];
-    CK_BYTE resp[blobsize];
+    CK_BYTE req[BLOBSIZE];
+    CK_BYTE resp[BLOBSIZE];
     CK_LONG req_len;
     size_t resp_len;
-    struct ep11_admresp rb;
-    struct ep11_admresp lrb;
-    ep11_target_t target;
+    struct XCPadmresp rb;
+    struct XCPadmresp lrb;
+    ep11_target_t target_list;
+    struct XCP_Module module;
+    target_t target = XCP_TGT_INIT;
     CK_RV rc;
     CK_BYTE name[256];
-
-    ep11_opaque *op_old = (ep11_opaque *) old;
-    ep11_opaque op_new;
-
+    unsigned char blob[BLOBSIZE];
+    CK_ULONG blob_len;
 
     CK_ATTRIBUTE opaque_template[] = {
-        {CKA_IBM_OPAQUE, &op_new, sizeof(op_new)}
+        { CKA_IBM_OPAQUE, blob, BLOBSIZE }
     };
 
     CK_ATTRIBUTE name_template[] = {
@@ -110,68 +119,126 @@ static int reencrypt(CK_SESSION_HANDLE session, CK_ULONG obj, CK_BYTE *old)
 
     memset(&rb, 0, sizeof(rb));
     memset(&lrb, 0, sizeof(lrb));
-    memset(&target, 0, sizeof(target));
-    target.length = 1;
 
-    target.apqns[0] = adapter;
-    target.apqns[1] = domain;
+    if (_m_add_module != NULL) {
+        memset(&module, 0, sizeof(module));
+        module.version = lib_version.major >= 3 ? XCP_MOD_VERSION_2
+                                                : XCP_MOD_VERSION_1;
+        module.flags = XCP_MFL_MODULE;
+        module.module_nr = adapter;
+        XCPTGTMASK_SET_DOM(module.domainmask, domain);
+        rc = _m_add_module(&module, &target);
+        if (rc != 0)
+            return CKR_FUNCTION_FAILED;
+    } else {
+        /* Fall back to old target handling */
+        memset(&target_list, 0, sizeof(ep11_target_t));
+        target_list.length = 1;
+        target_list.apqns[0] = adapter;
+        target_list.apqns[1] = domain;
+        target = (target_t)&target_list;
+    }
+
     rb.domain = domain;
     lrb.domain = domain;
 
-    fprintf(stderr, "going to reencrpyt key %lx with blob len %lx %s\n", obj,
-            op_old->blob_size, name);
-    resp_len = blobsize;
+    fprintf(stderr, "going to reencrpyt key %lx with blob len %lx: '%s'\n", obj,
+            old_len, name);
+    resp_len = BLOBSIZE;
 
-    req_len = _ep11a_cmdblock(req, blobsize, EP11_ADM_REENCRYPT, &rb,
-                              NULL, op_old->blob, op_old->blob_size);
+    req_len = _xcpa_cmdblock(req, BLOBSIZE, XCP_ADM_REENCRYPT, &rb,
+                              NULL, old, old_len);
 
     if (req_len < 0) {
         fprintf(stderr, "reencrypt cmd block construction failed\n");
-        return -2;
+        rc = -2;
+        goto out;
     }
 
     rc = _m_admin(resp, &resp_len, NULL, 0, req, req_len, NULL, 0,
-                  (unsigned long long) &target);
+                  target);
 
     if (rc != CKR_OK || resp_len == 0) {
-        fprintf(stderr, "reencryption failed %lx %ld\n", rc, req_len);
-        return -3;
+        fprintf(stderr, "reencryption failed: %lx %ld\n", rc, req_len);
+        rc = -3;
+        goto out;
     }
 
-    if (_ep11a_internal_rv(resp, resp_len, &lrb, &rc) < 0) {
-        fprintf(stderr, "reencryption response malformed %lx\n", rc);
-        return -4;
+    if (_xcpa_internal_rv(resp, resp_len, &lrb, &rc) < 0) {
+        fprintf(stderr, "reencryption response malformed: %lx\n", rc);
+        rc = -4;
+        goto out;
     }
 
-    if (op_old->blob_size != lrb.pllen) {
-        fprintf(stderr, "reencryption blob size changed %lx %lx %lx %lx\n",
-                op_old->blob_size, lrb.pllen, resp_len, req_len);
-        return -5;
+    if (rc != 0) {
+        fprintf(stderr, "reencryption failed: %lx\n", rc);
+        rc = -7;
+        goto out;
     }
 
-    memset(&op_new, 0, sizeof(op_new));
-    op_new.blob_id = op_old->blob_id;
-    op_new.blob_size = op_old->blob_size;
-    memcpy(op_new.blob, lrb.payload, op_new.blob_size);
+    if (old_len != lrb.pllen) {
+        fprintf(stderr, "reencryption blob size changed: %lx %lx %lx %lx\n",
+                old_len, lrb.pllen, resp_len, req_len);
+        rc = -5;
+        goto out;
+    }
+
+    memset(blob, 0, sizeof(blob));
+    blob_len = old_len;
+    memcpy(blob, lrb.payload, blob_len);
+    opaque_template[0].ulValueLen = blob_len;
 
     rc = funcs->C_SetAttributeValue(session, key_store[obj], opaque_template,
                                     1);
     if (rc != CKR_OK) {
         fprintf(stderr,
-                "reencryption C_SetAttributeValue failed obj %lx %s rc %lx\n",
+                "reencryption C_SetAttributeValue failed: obj %lx '%s' rc: %lx\n",
                 obj, name, rc);
-        return -6;
+        rc = -6;
+        goto out;
     }
 
-    fprintf(stderr, "reencryption success obj %lx %s\n", obj, name);
-    return 0;
+    fprintf(stderr, "reencryption success obj: %lx '%s:\n", obj, name);
+
+out:
+    if (_m_rm_module != NULL)
+        _m_rm_module(&module, target);
+    return rc;
+}
+
+static CK_RV get_ep11_library_version(CK_VERSION *lib_version)
+{
+    unsigned int host_version;
+    CK_ULONG version_len = sizeof(host_version);
+    CK_RV rc;
+
+    rc = _m_get_xcp_info(&host_version, &version_len,
+                         CK_IBM_XCPHQ_VERSION, 0, 0);
+    if (rc != CKR_OK) {
+        fprintf(stderr, "_m_get_xcp_info (HOST) failed: rc=0x%lx\n", rc);
+        return rc;
+    }
+    lib_version->major = (host_version & 0x00FF0000) >> 16;
+    lib_version->minor = host_version & 0x000000FF0000;
+    /*
+     * EP11 host library < v2.0 returns an invalid version (i.e. 0x100). This
+     * can safely be treated as version 1.0
+     */
+    if (lib_version->major == 0) {
+        lib_version->major = 1;
+        lib_version->minor = 0;
+    }
+
+    return CKR_OK;
 }
 
 
 static int check_card_status()
 {
     CK_RV rc;
-    ep11_target_t target;
+    ep11_target_t target_list;
+    struct XCP_Module module;
+    target_t target = XCP_TGT_INIT;
     CK_IBM_DOMAIN_INFO dinf;
     CK_ULONG dinf_len = sizeof(dinf);
 
@@ -180,18 +247,33 @@ static int check_card_status()
         return -1;
     }
 
-    target.format = 0;
-    target.length = 1;
-    target.apqns[0] = adapter;
-    target.apqns[1] = domain;
+    if (_m_add_module != NULL) {
+        memset(&module, 0, sizeof(module));
+        module.version = lib_version.major >= 3 ? XCP_MOD_VERSION_2
+                                                : XCP_MOD_VERSION_1;
+        module.flags = XCP_MFL_MODULE;
+        module.module_nr = adapter;
+        XCPTGTMASK_SET_DOM(module.domainmask, domain);
+        rc = _m_add_module(&module, &target);
+        if (rc != 0)
+            return CKR_FUNCTION_FAILED;
+    } else {
+        /* Fall back to old target handling */
+        memset(&target_list, 0, sizeof(ep11_target_t));
+        target_list.length = 1;
+        target_list.apqns[0] = adapter;
+        target_list.apqns[1] = domain;
+        target = (target_t)&target_list;
+    }
 
-    rc = _m_get_ep11_info((CK_VOID_PTR) &dinf, &dinf_len,
-                          CK_IBM_EP11Q_DOMAIN, 0, (unsigned long long) &target);
+    rc = _m_get_xcp_info((CK_VOID_PTR) &dinf, &dinf_len,
+                         CK_IBM_XCPQ_DOMAIN, 0, target);
 
     if (rc != CKR_OK) {
-        fprintf(stderr, "m_get_ep11_info rc 0x%lx, valid apapter/domain "
+        fprintf(stderr, "m_get_xcp_info rc 0x%lx, valid apapter/domain "
                 "0x%02lx/%ld?.\n", rc, adapter, domain);
-        return -1;
+        rc = -1;
+        goto out;
     }
 
     if (CK_IBM_DOM_COMMITTED_NWK & dinf.flags) {
@@ -201,11 +283,15 @@ static int check_card_status()
         fprintf(stderr,
                 "Card ID 0x%02lx, domain ID %ld has no committed pending WK\n",
                 adapter, domain);
-        return -1;
+        rc = -1;
+        goto out;
     }
 
-    return 0;
+out:
+    if (_m_rm_module != NULL)
+         _m_rm_module(&module, target);
 
+    return rc;
 }
 
 
@@ -388,6 +474,61 @@ static int do_ParseArgs(int argc, char **argv)
     return 1;
 }
 
+#ifdef EP11_HSMSIM
+#define DLOPEN_FLAGS        RTLD_GLOBAL | RTLD_NOW | RTLD_DEEPBIND
+#else
+#define DLOPEN_FLAGS        RTLD_GLOBAL | RTLD_NOW
+#endif
+
+static void *ep11_load_host_lib()
+{
+    void *lib_ep11;
+    char *ep11_lib_name;
+    char *errstr;
+
+    ep11_lib_name = getenv(EP11SHAREDLIB_NAME);
+    if (ep11_lib_name != NULL) {
+        lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
+
+        if (lib_ep11 == NULL) {
+            errstr = dlerror();
+            fprintf(stderr, "Error loading shared library '%s' [%s]\n",
+                    ep11_lib_name, errstr);
+            return NULL;
+        }
+        return lib_ep11;
+    }
+
+    ep11_lib_name = EP11SHAREDLIB_V3;
+    lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
+
+    if (lib_ep11 == NULL) {
+        /* Try version 2 instead */
+        ep11_lib_name = EP11SHAREDLIB_V2;
+        lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
+    }
+
+    if (lib_ep11 == NULL) {
+        /* Try version 1 instead */
+        ep11_lib_name = EP11SHAREDLIB_V1;
+        lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
+    }
+
+    if (lib_ep11 == NULL) {
+        /* Try unversioned library instead */
+        ep11_lib_name = EP11SHAREDLIB;
+        lib_ep11 = dlopen(ep11_lib_name, DLOPEN_FLAGS);
+    }
+
+    if (lib_ep11 == NULL) {
+        errstr = dlerror();
+        fprintf(stderr, "Error loading shared library '%s[.3|.2|.1]' [%s]\n",
+                EP11SHAREDLIB, errstr);
+        return NULL;
+    }
+
+    return lib_ep11;
+}
 
 int main(int argc, char **argv)
 {
@@ -407,25 +548,40 @@ int main(int argc, char **argv)
     }
 
     /* dynamically load in the ep11 shared library */
-    lib_ep11 = dlopen(EP11SHAREDLIB, RTLD_GLOBAL | RTLD_NOW);
-    if (!lib_ep11) {
-        fprintf(stderr, "ERROR loading shared lib '%s' [%s]", EP11SHAREDLIB,
-                dlerror());
+    lib_ep11 = ep11_load_host_lib();
+    if (!lib_ep11)
         return CKR_FUNCTION_FAILED;
-    }
 
-    *(void **)(&_m_get_ep11_info) = dlsym(lib_ep11, "m_get_ep11_info");
-    *(void **)(&_ep11a_cmdblock) = dlsym(lib_ep11, "ep11a_cmdblock");
+    *(void **)(&_xcpa_cmdblock) = dlsym(lib_ep11, "xcpa_cmdblock");
+    if (_xcpa_cmdblock == NULL)
+        *(void **)(&_xcpa_cmdblock) = dlsym(lib_ep11, "ep11a_cmdblock");
     *(void **)(&_m_admin) = dlsym(lib_ep11, "m_admin");
-    *(void **)(&_ep11a_internal_rv) = dlsym(lib_ep11, "ep11a_internal_rv");
+    *(void **)(&_xcpa_internal_rv) = dlsym(lib_ep11, "xcpa_internal_rv");
+    if (_xcpa_internal_rv == NULL)
+        *(void **)(&_xcpa_internal_rv) = dlsym(lib_ep11, "ep11a_internal_rv");
+    *(void **)(&_m_get_xcp_info) = dlsym(lib_ep11, "m_get_xcp_info");
+    if (_m_get_xcp_info == NULL)
+        *(void **)(&_m_get_xcp_info) = dlsym(lib_ep11, "m_get_ep11_info");
 
-    if (!_m_get_ep11_info || !_ep11a_cmdblock ||
-        !_m_admin || !_ep11a_internal_rv) {
+    if (!_m_get_xcp_info || !_xcpa_cmdblock ||
+        !_m_admin || !_xcpa_internal_rv) {
         fprintf(stderr, "ERROR getting function pointer from shared lib '%s'",
                 EP11SHAREDLIB);
         return CKR_FUNCTION_FAILED;
     }
 
+    /*
+     * The following are only available since EP11 host library version 2.
+     * Ignore if they fail to load, the code will fall back to the old target
+     * handling in this case.
+     */
+    *(void **)(&_m_add_module) = dlsym(lib_ep11, "m_add_module");
+    *(void **)(&_m_rm_module) = dlsym(lib_ep11, "m_rm_module");
+    if (_m_add_module == NULL || _m_rm_module == NULL) {
+        _m_add_module = NULL;
+        _m_rm_module = NULL;
+    }
+
     printf("Using slot #%lu...\n\n", SLOT_ID);
 
     rc = do_GetFunctionList();
@@ -476,6 +632,10 @@ int main(int argc, char **argv)
         return rc;
     }
 
+    rc = get_ep11_library_version(&lib_version);
+    if (rc != CKR_OK)
+        return rc;
+
     if (check_card_status() != 0)
         return 1;
 
@@ -525,7 +685,8 @@ int main(int argc, char **argv)
                     return rc;
                 } else {
                     if (reencrypt(session, obj,
-                                  (CK_BYTE *) opaque_template[0].pValue) != 0) {
+                                  (CK_BYTE *) opaque_template[0].pValue,
+                                  opaque_template[0].ulValueLen) != 0) {
                         /* reencrypt failed */
                         return -1;
                     }
